"""
pygene is a library for genetic algorithms in python

It aims to be very simple to use, and suitable for people
new to genetic algorithms.
"""

version = "0.3.0"

__all__ = [
    'gene', 'gamete', 'organism', 'population', 'xmlio', 'prog',
    'config'
    ]

